"""
Tests Non-Repairable Optimal Replacement Time algorithms.

Uses pytest fixtures located in conftest.py in the tests/ directory.
"""
# import numpy as np
# import pytest
# from surpyval.renewal import Crow

# from repyability.repairable import Repairable

# https://reliawiki.org/index.php/Repairable_Systems_Analysis
# model = Crow.from_params([1./2.1211e-5, 1.4738])
